from uuid import uuid4

from eoxserver.contrib import gdal, osr
from eoxserver.resources.coverages import crss


def create_mem_ds(width, height, data_type):
    driver = gdal.GetDriverByName('MEM')
    return driver.Create('', width, height, 1, data_type)


def warp_fields(coverages, field_name, bbox, crs, width, height):
    driver = gdal.GetDriverByName('MEM')
    out_ds = driver.Create(
        '',
        width,
        height,
        1,
        coverages[0].range_type.get_field(field_name).data_type
    )

    out_ds.SetGeoTransform([
        bbox[0],
        (bbox[2] - bbox[0]) / width,
        0,
        bbox[3],
        0,
        -(bbox[3] - bbox[1]) / height,
    ])
    epsg = crss.parseEPSGCode(crs, [crss.fromShortCode])
    sr = osr.SpatialReference()
    sr.ImportFromEPSG(epsg)

    out_ds.SetProjection(sr.ExportToWkt())

    for coverage in coverages:
        location = coverage.get_location_for_field(field_name)
        band_index = coverage.get_band_index_for_field(field_name)

        orig_ds = gdal.open_with_env(location.path, location.env)

        vrt_filename = None
        if orig_ds.RasterCount > 1:
            vrt_filename = '/vsimem/' + uuid4().hex
            gdal.BuildVRT(vrt_filename, orig_ds, bandList=[band_index])
            ds = gdal.Open(vrt_filename)
        else:
            ds = orig_ds

        gdal.Warp(out_ds, ds)
        ds = None

        if vrt_filename:
            gdal.Unlink(vrt_filename)

    band = out_ds.GetRasterBand(1)
    return band.ReadAsArray()



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from lxml import etree

from eoxserver.core import Component, implements
from eoxserver.core.config import get_eoxserver_config
from eoxserver.core.util.timetools import isoformat
from eoxserver.contrib.mapserver import create_request, Map, Layer
from eoxserver.resources.coverages import crss
from eoxserver.render.coverage.objects import Coverage
from eoxserver.services.mapserver.wcs.base_renderer import BaseRenderer
from eoxserver.services.ows.common.config import CapabilitiesConfigReader
from eoxserver.services.ows.wcs.interfaces import (
    WCSCapabilitiesRendererInterface
)
from eoxserver.services.ows.version import Version
from eoxserver.services.result import result_set_from_raw_data, get_content_type, ResultBuffer
from eoxserver.services.urls import get_http_service_url


class MapServerWCSCapabilitiesRenderer(BaseRenderer):
    """ WCS Capabilities renderer implementation using MapServer.
    """
    implements(WCSCapabilitiesRendererInterface)

    versions = (Version(1, 0), Version(1, 1))

    def supports(self, params):
        return params.version in self.versions

    def render(self, params):
        conf = CapabilitiesConfigReader(get_eoxserver_config())

        http_service_url = get_http_service_url(params.http_request)

        map_ = Map()
        map_.setMetaData({
            "enable_request": "*",
            "onlineresource": http_service_url,
            "service_onlineresource": conf.onlineresource,
            "updateSequence": conf.update_sequence,
            "name": conf.name,
            "title": conf.title,
            "label": conf.title,
            "abstract": conf.abstract,
            "accessconstraints": conf.access_constraints,
            "addresstype": "",
            "address": conf.delivery_point,
            "stateorprovince": conf.administrative_area,
            "city": conf.city,
            "postcode": conf.postal_code,
            "country": conf.country,
            "contactelectronicmailaddress": conf.electronic_mail_address,
            "contactfacsimiletelephone": conf.phone_facsimile,
            "contactvoicetelephone": conf.phone_voice,
            "contactperson": conf.individual_name,
            "contactorganization": conf.provider_name,
            "contactposition": conf.position_name,
            "role": conf.role,
            "hoursofservice": conf.hours_of_service,
            "contactinstructions": conf.contact_instructions,
            "fees": conf.fees,
            "keywordlist": ",".join(conf.keywords),
            "formats": " ".join([f.wcs10name for f in self.get_wcs_formats()]),
            "srs": " ".join(crss.getSupportedCRS_WCS(format_function=crss.asShortCode)),
        }, namespace="ows")
        map_.setProjection("EPSG:4326")

        for outputformat in self.get_all_outputformats(False):
            map_.appendOutputFormat(outputformat)

        for coverage in params.coverages:
            layer = Layer(coverage.identifier)

            render_coverage = Coverage.from_model(coverage)
            layer.setProjection(render_coverage.grid.spatial_reference.proj)
            extent = render_coverage.extent
            size = render_coverage.size
            resolution = ((extent[2] - extent[0]) / float(size[0]),
                          (extent[1] - extent[3]) / float(size[1]))

            layer.setExtent(*extent)
            layer.setMetaData({
                "title": coverage.identifier,
                "label": coverage.identifier,
                "extent": "%.10g %.10g %.10g %.10g" % extent,
                "resolution": "%.10g %.10g" % resolution,
                "size": "%d %d" % size,
                "formats": " ".join([f.wcs10name for f in self.get_wcs_formats()]),
                "srs": " ".join(crss.getSupportedCRS_WCS(format_function=crss.asShortCode)),
            }, namespace="wcs")

            map_.insertLayer(layer)

        request = create_request(params)
        request.setParameter("version", params.version)
        raw_result = map_.dispatch(request)
        result = result_set_from_raw_data(raw_result)
        xml_result = etree.fromstring(result[0].data)

        for elem in xml_result.xpath('//*[local-name() = "metadataLink"]'):
            elem.getparent().remove(elem)

        # Add CQL parameter to GetCapabilities operation
        for elem in xml_result.xpath('//*[local-name() = "Operation"][@name = "GetCapabilities"]'):
            ows = elem.nsmap['ows']
            param = etree.SubElement(elem, '{%s}Parameter' % ows)
            param.attrib['name'] = 'cql'
            etree.SubElement(param, '{%s}AnyValue' % ows)

        xml_result_data = etree.tostring(xml_result, pretty_print=True, encoding='UTF-8', xml_declaration=True)

        result[0] = ResultBuffer(xml_result_data, result[0].content_type)
        return result



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


import logging

from eoxserver.resources.coverages import models
from eoxserver.core.decoders import InvalidParameterException
from eoxserver.services.subset import Trim, Slice
from eoxserver.services.ows.wms.exceptions import LayerNotDefined
from eoxserver.services.ows.wms.parsing import (parse_bbox, parse_time, int_or_str)


logger = logging.getLogger(__name__)


def lookup_layers(layers, subsets, suffixes=None):
    """ Performs a layer lookup for the given layer names. Applies the given
        subsets and looks up all layers with the given suffixes. Returns a
        hierarchy of ``LayerSelection`` objects.
    """
    suffix_related_ids = {}
    root_group = LayerSelection(None)
    suffixes = suffixes or (None,)
    logger.debug(str(suffixes))

    for layer_name in layers:
        for suffix in suffixes:
            if not suffix:
                identifier = layer_name
            elif layer_name.endswith(suffix):
                identifier = layer_name[:-len(suffix)]
            else:
                continue

            # TODO: nasty, nasty bug... dunno where
            eo_objects = models.EOObject.objects.filter(
                identifier=identifier
            )
            if len(eo_objects):
                eo_object = eo_objects[0]
                break
        else:
            raise LayerNotDefined(layer_name)

        if models.iscollection(eo_object):
            # recursively iterate over all sub-collections and collect all
            # coverages

            used_ids = suffix_related_ids.setdefault(suffix, set())

            def recursive_lookup(collection, suffix, used_ids, subsets):
                # get all EO objects related to this collection, excluding
                # those already searched
                eo_objects = models.EOObject.objects.filter(
                    collections__in=[collection.pk]
                ).exclude(
                    pk__in=used_ids
                ).order_by("begin_time", "end_time")
                # apply subsets
                eo_objects = subsets.filter(eo_objects)

                selection = LayerSelection()

                # append all retrived EO objects, either as a coverage of
                # the real type, or as a subgroup.
                for eo_object in eo_objects:
                    used_ids.add(eo_object.pk)

                    if models.iscoverage(eo_object):
                        selection.append(eo_object.cast(), eo_object.identifier)
                    elif models.iscollection(eo_object):
                        selection.extend(recursive_lookup(
                            eo_object, suffix, used_ids, subsets
                        ))
                    else:
                        pass

                return selection

            root_group.append(
                LayerSelection(
                    eo_object, suffix,
                    recursive_lookup(eo_object, suffix, used_ids, subsets)
                )
            )

        elif models.iscoverage(eo_object):
            # Add a layer selection for the coverage with the suffix
            selection = LayerSelection(None, suffix=suffix)
            if subsets.matches(eo_object):
                selection.append(eo_object.cast(), eo_object.identifier)
            else:
                selection.append(None, eo_object.identifier)

            root_group.append(selection)

    return root_group


class LayerSelection(list):
    """ Helper class for hierarchical layer selections.
    """
    def __init__(self, collection=None, suffix=None, iterable=None):
        self.collection = collection
        self.suffix = suffix
        if iterable:
            super(LayerSelection, self).__init__(iterable)

    def __contains__(self, eo_object):
        for item in self:
            try:
                if eo_object in item:
                    return True
            except TypeError:
                pass

            try:
                if eo_object == item[0]:
                    return True
            except IndexError:
                pass

        return False

    def append(self, eo_object_or_selection, name=None):
        if isinstance(eo_object_or_selection, LayerSelection):
            super(LayerSelection, self).append(eo_object_or_selection)
        else:
            super(LayerSelection, self).append((eo_object_or_selection, name))

    def walk(self, depth_first=True):
        """ Yields four-tuples (collections, coverage, name, suffix).
        """

        collection = (self.collection,) if self.collection else ()

        for item in self:
            try:
                for collections, eo_object, name, suffix in item.walk():
                    yield (
                        collection + collections,
                        eo_object, name,
                        suffix or self.suffix
                    )

            except AttributeError:
                yield collection, item[0], item[1], self.suffix

        if not self:
            yield collection, None, None, self.suffix



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

"""
 This module contains format handling utilities.
"""
#-------------------------------------------------------------------------------

import re
import sys
import imp
import logging
import os.path

from django.conf import settings

from eoxserver.contrib import gdal
from eoxserver.core.config import get_eoxserver_config
from eoxserver.core.decoders import config, typelist, strip

try:
    # Python 2
    xrange
except NameError:
    # Python 3, xrange is now named range
    xrange = range

logger = logging.getLogger(__name__)

#-------------------------------------------------------------------------------


class FormatRegistryException(Exception):
    pass


class Format(object):

    """
    Format record class.
    The class is rather structure with read-only properties (below).
    The class implements ``__str__()`` and ``__eq__()`` methods.
    """

    def __init__(self, mime_type, driver, extension, is_writeable):
        self.__mimeType = mime_type
        self.__driver = driver
        self.__defaultExt = extension
        self.__isWriteable = is_writeable

    mimeType = property(lambda self: self.__mimeType, doc="MIME-type")
    driver = property(lambda self: self.__driver, doc="library/driver identifier")
    defaultExt = property(lambda self: self.__defaultExt, doc="default extension (including dot)")
    isWriteable = property(lambda self: self.__isWriteable, doc="boolean flag indicating that output can be produced")

    @property
    def wcs10name(self):
        """ get WCS 1.0 format name """
        if self.driver.startswith("GDAL/"):
            s = self.driver.split('/')[1]
            if s == "GTiff":
                s = "GeoTIFF"
        else:
            s = self.driver.replace("/", ":")
        return s

    def __str__(self):
        return "%s,%s,%s #%s" % (
            self.mimeType, self.driver, self.defaultExt,
            "rw" if self.isWriteable else "ro"
        )

    def __eq__(self, other):
        try:
            return (
                self.mimeType == other.mimeType and
                self.driver == other.driver and
                self.defaultExt == other.defaultExt
            )
        except AttributeError:
            return False

    def __hash__(self):
        return hash((self.mimeType, self.driver, self.defaultExt, self.isWriteable))

# ------------------------------------------------------------------------------

__FORMAT_REGISTRY = None


class FormatRegistry(object):
    """
    The :class:`FormatRegistry` class represents cofiguration of file supported
    formats and of the auxiliary methods. The formats' configuration relies
    on two configuration files:

    * the default formats' configuration (``eoxserver/conf/default_formats.conf``)
    * the optional instance configuration (``conf/format.conf`` in the instance
      directory)

    Configuration values are read from these files.
    """

    #---------------------------------------------------------------------------

    def __init__(self, config):
        # get path to EOxServer installation
        path_eoxs = self.__get_path_eoxs()

        # default formats' configuration
        path_formats_def = os.path.join(path_eoxs, "conf", "default_formats.conf")

        if not os.path.exists(path_formats_def):
            # try alternative location
            path_formats_def = os.path.join(
                sys.prefix, "eox_server", "conf", "default_formats.conf"
            )

            if not os.path.exists(path_formats_def):
                # failed to read the file
                raise FormatRegistryException(
                    "Cannot find the default file formats' configuration file."
                )

        # optional formats' configuration
        path_formats_opt = os.path.join(
            settings.PROJECT_DIR, "conf", "formats.conf"
        )

        if not os.path.exists(path_formats_opt):
            path_formats_opt = None  # no user defined formats' configuration
            logger.debug(
                "Optional, user-defined file formats' specification not found. "
                "Only the installation defaults will be used."
            )

        # load the formats' configuaration
        self.__load_formats(path_formats_def, path_formats_opt)

        # parse the config options
        self.__parse_config(config)

    # --------------------------------------------------------------------------
    # getters

    def getFormatsAll(self):
        """ Get list of all registered formats """

        return self.__mime2format.values()

    def getFormatsByDriver(self, driver_name):
        """
        Get format records for the given GDAL driver name.
        In case of no match empty list is returned.
        """

        return self.__driver2format.get(valDriver(driver_name), [])

    def getFormatsByWCS10Name(self, wcs10name):
        """
        Get format records for the given GDAL driver name. In case of no
        match an empty list is returned.
        """

        # convert WCS 1.0 format name to driver name
        if ":" in wcs10name:
            driver_name = wcs10name.replace(":", "/")
        else:
            if "GeoTIFF" == wcs10name:
                wcs10name = "GTiff"
            driver_name = "GDAL/%s" % wcs10name

        return self.getFormatsByDriver(driver_name)

    def getFormatByMIME(self, mime_type):
        """
        Get format record for the given MIME type.
        In case of no match None is returned.
        """
        return self.__mime2format.get(valMimeType(mime_type), None)

    # --------------------------------------------------------------------------
    # OWS specific getters

    def getSupportedFormatsWCS(self):
        """
            Get list of formats to be announced as supported WCS formats.

            The the listed formats must be:
            * defined in EOxServers configuration (section "services.ows.wcs", item "supported_formats") 
            * defined in the formats' configuration ("default_formats.conf" or "formats.conf")   
            * supported by the used GDAL installation
        """
        return self.__wcs_supported_formats

    def getSupportedFormatsWMS(self):
        """
            Get list of formats to be announced as supported WMS formats.

            The the listed formats must be:
            * defined in EOxServers configuration (section "services.ows.wms", item "supported_formats") 
            * defined in the formats' configuration ("default_formats.conf" or "formats.conf")   
            * supported by the used GDAL installation
        """
        return self.__wms_supported_formats

    def mapSourceToNativeWCS20(self, format):
        """ Map source format to WCS 2.0 native format.

        Both the input and output shall be instances of :class:`Formats` class.
        The input format can be obtained, e.g., by the `getFormatByDriver` or `getFormatByMIME` 
        method.

        To force the default native format use None as the source format.

        The format mapping follows these rules:

        1. Mapping based on the explicite rules is applied if possible (defined in EOxServers
           configuration, section "services.ows.wcs20", item "source_to_native_format_map").
           If there is no mapping available the source format is kept.
        2. If the format resulting from step 1 is not a writable GDAL format or
           it is not among the supported WCS formats than it is
           replaced by the default native format (defined in EOxServers
           configuration, section "services.ows.wcs20", item "default_native_format"). 
           In case of writable GDAL format, the result of step 1 is returned.
        """

        # 1. apply mapping
        format = self.__wcs20_format_mapping.get(format, format)

        # 2. fallback to default
        if format is None or not format.isWriteable \
                or format not in self.getSupportedFormatsWCS():

            format = self.__wcs20_def_native_format

        return format

    def getDefaultNativeFormat(self):
        """ Get default nativeFormat as defined in section 'services.ows.wcs20'.
        """
        return self.__wcs20_def_native_format

    # --------------------------------------------------------------------------
    # loading of configuration - private auxiliary subroutines

        # parse the config options
    def __parse_config(self, config):
        """
        Parse the EOxServer configuration.
        """

        reader = FormatConfigReader(config)

        #  WMS and WCS suported formats

        nonNone = lambda v: (v is not None)
        self.__wms_supported_formats = list(filter(nonNone, map(self.getFormatByMIME, reader.supported_formats_wms)))
        self.__wcs_supported_formats = list(filter(nonNone, map(self.getFormatByMIME, reader.supported_formats_wcs)))

        #  WCS 2.0.1 source to native format mapping

        tmp = self.getFormatByMIME(reader.default_native_format)

        self.__wcs20_def_native_format = tmp

        if tmp is None or tmp not in self.getSupportedFormatsWCS():
            raise ValueError(
                "Invalid value of configuration option 'services.ows.wcs20' "
                "'default_native_format'! value=\"%s\"" % src
            )

        tmp = reader.source_to_native_format_map
        tmp = list(map(lambda m: self.getFormatByMIME(m.strip()), list(tmp.split(','))))
        tmp = [(tmp[i], tmp[i + 1]) for i in xrange(0, (len(tmp) >> 1) << 1, 2)]
        tmp = list(filter(lambda p: list(p)[0] is not None and list(p)[1] is not None, tmp))
        self.__wcs20_format_mapping = dict(tmp)

    def __load_formats(self, path_formats_def, path_formats_opt):
        """
        Load and parse the formats' configuration.
        """

        # register GDAL drivers
        gdal.AllRegister()

        # reset iternall format storage
        self.__driver2format = {}
        self.__mime2format = {}

        # read default configuration
        logger.debug("Loading formats' configuration from: %s" % path_formats_def)
        with open(path_formats_def) as f:
            for ln, line in enumerate(f):
                self.__parse_line(line, path_formats_def, ln + 1)

        # read the optional configuration
        if path_formats_opt:
            logger.debug("Loading formats' configuration from: %s" % path_formats_opt)
            with open(path_formats_opt) as f:
                for ln, line in enumerate(f):
                    self.__parse_line(line, path_formats_opt, ln + 1)

        # finalize format specification
        self.__postproc_formats()

    def __postproc_formats(self):
        """
        Postprocess format specificaions after the loading was finished.
        """

        for frec in list(self.__mime2format.values()):
            # driver to format dictionary
            if frec.driver in self.__driver2format:
                self.__driver2format.append(frec)
            else:
                self.__driver2format[frec.driver] = [frec]

    def __parse_line(self, line, fname, lnum):
        """
        Parse single line of configuration.
        """

        # parse line
        try:
            line = line.partition("#")[0].strip()  # strip comments and white characters 

            if not line:
                return

            (mime_type, driver, extension) = line.split(',')

            mime_type = valMimeType(mime_type.strip())
            driver = valDriver(driver.strip())
            extension = extension.strip()

            if None in (driver, mime_type):
                raise ValueError("Invalid input format specification \"%s\"!" % line)

            # check the check the driver
            backend, _, ldriver = driver.partition("/")

            # no-other backend than GDAL currently supported

            if backend == "GDAL":
                gdriver = gdal.GetDriverByName(ldriver)

                if gdriver is None:
                    raise ValueError("Invalid GDAL driver \"%s\"!" % driver)

                #get the writebility
                is_writeable = (gdriver.GetMetadataItem("DCAP_CREATECOPY") == "YES")

            else:

                raise ValueError("Invalid driver backend \"%s\"!" % driver)

            # create new format record
            frec = Format(mime_type, driver, extension, is_writeable)

            # store format record
            self.__mime2format[mime_type] = frec

            logger.debug("Adding new file format: %s" % frec)

        except Exception as e:
            logger.warning(
                "%s:%i Invalid file format specification! Line ignored! "
                "line=\"%s\" message=\"%s\"" % (
                    fname, lnum, line, e
                )
            )

    def __get_path_eoxs(self):
        """
        Get path to the EOxServer installation.
        """

        try:
            return imp.find_module("eoxserver")[1]
        except ImportError:
            raise FormatRegistryException(
                "Filed to find the 'eoxserver' module! Check your modules' path!"
            )


class FormatConfigReader(config.Reader):
    section = "services.ows.wms"
    supported_formats_wms = config.Option("supported_formats", typelist(strip, ","))

    section = "services.ows.wcs"
    supported_formats_wcs = config.Option("supported_formats", typelist(strip, ","))

    section = "services.ows.wcs20"
    default_native_format = config.Option(type=strip)
    source_to_native_format_map = config.Option()


#-------------------------------------------------------------------------------
# regular expression validators

#: MIME-type regular expression validator (compiled reg.ex. pattern)
_gerexValMime = re.compile("^[\w][-\w]*/[\w][-+\w]*(;[-\w]*=[-\w]*)*$")

#: library driver regular expression validator (compiled reg.ex. pattern)
_gerexValDriv = re.compile("^[\w][-\w]*/[\w][-\w]*$")


def valMimeType(string):
    """
    MIME type reg.ex. validator. If pattern not matched 'None' is returned
    otherwise the input is returned.
    """
    rv = string if _gerexValMime.match(string) else None
    if None is rv:
        logger.warning("Invalid MIME type \"%s\"." % string)
    return rv


def valDriver(string):
    """
    Driver identifier reg.ex. validator. If pattern not matched 'None' is returned 
    otherwise the input is returned.
    """
    rv = string if _gerexValDriv.match(string) else None
    if None is rv:
        logger.warning("Invalid driver's identifier \"%s\"." % string)
    return rv

#-------------------------------------------------------------------------------
# public API


def getFormatRegistry():
    """
        Get initialised instance of the FormatRegistry class.
        This is the preferable way to get the Format Registry.
    """

    global __FORMAT_REGISTRY

    if __FORMAT_REGISTRY is None:

        logger.debug(" --- getFormatRegistry() --- ")
        logger.debug(repr(__FORMAT_REGISTRY))
        logger.debug(repr(_gerexValMime))
        logger.debug(repr(_gerexValDriv))

        # load configuration if not already loaded
        __FORMAT_REGISTRY = FormatRegistry(get_eoxserver_config())

        logger.debug(repr(__FORMAT_REGISTRY))

    return __FORMAT_REGISTRY

#-------------------------------------------------------------------------------



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Stephan Krause <stephan.krause@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#          Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import logging
import traceback
from optparse import OptionValueError

import django
from django.db import transaction
from django.core.management.base import CommandParser


logger = logging.getLogger(__name__)


def _variable_args_cb(option, opt_str, value, parser):
    """ Helper function for optparse module. Allows
        variable number of option values when used
        as a callback.
    """
    args = []
    for arg in parser.rargs:
        if not arg.startswith('-'):
            args.append(arg)
        else:
            break

    del parser.rargs[:len(args)]

    if getattr(parser.values, option.dest):
        args.extend(getattr(parser.values, option.dest))
    setattr(parser.values, option.dest, args)


class StringFormatCallback(object):
    """ Small helper class to supply a variable number of arguments to a
    callback function and store the resulting value in the `dest` field of the
    parser.
    """

    def __init__(self, callback):
        self.callback = callback

    def __call__(self, option, opt_str, value, parser):
        args = []
        for arg in parser.rargs:
            if not arg.startswith('-'):
                args.append(arg)
            else:
                del parser.rargs[:len(args)]
                break

        try:
            setattr(parser.values, option.dest, self.callback(" ".join(args)))
        except ValueError as e:
            raise OptionValueError(str(e))


class CommandOutputMixIn(object):
    """ Helper mix-in class to ease the handling of user message reporting.
    """

    def print_err(self, msg):
        " Print an error message which is both logged and written to stderr. "

        logger.error(msg)
        self.stderr.write("ERROR: %s\n" % msg)

    def print_wrn(self, msg):
        """ Print a warning message, which is logged and posibly written to
            stderr, depending on the set verbosity.
        """

        logger.warning(msg)

        if 0 < max(0, getattr(self, "verbosity", 1)):
            self.stderr.write("WARNING: %s\n" % msg)

    def print_msg(self, msg, level=1):
        """ Print a basic message with a given level. The message is possibly
            logged and/or written to stderr depending on the verbosity setting.
        """

        # three basic level of info messages
        # level == 0 - always printed even in the silent mode - not recommended
        # level == 1 - normal info suppressed in silent mode
        # level >= 2 - debuging message (additional levels allowed)
        # messages ALWAYS logged (as either info or debug)

        level = max(0, level)
        verbosity = max(0, getattr(self, "verbosity", 1))

        # everything with level 2 or higher is DEBUG
        if level >= 2:
            prefix = "DEBUG"
            logger.debug(msg)

        # levels 0 (silent) and 1 (default-verbose)
        else:
            prefix = "INFO"
            logger.info(msg)

        if level <= verbosity:
            self.stdout.write("%s: %s\n" % (prefix, msg))

    def print_traceback(self, e, kwargs):
        """ Prints a traceback/stacktrace if the traceback option is set.
        """
        if kwargs.get("traceback", False):
            self.print_msg(traceback.format_exc())


def create_parser(cmd, kwargs):
    if django.VERSION[0] < 2:
        return CommandParser(cmd, **kwargs)
    else:
        return CommandParser(**kwargs)

class SubParserMixIn(object):
    def add_subparser(self, parser, name, *args, **kwargs):
        if not getattr(self, 'subparsers', None):
            self.subparsers = parser.add_subparsers(
                title="subcommands",
                parser_class=lambda **kw: create_parser(self, kw)
            )
        subparser = self.subparsers.add_parser(name, *args, **kwargs)
        subparser.set_defaults(subcommand=name)

        subparser.add_argument('--traceback', action="store_true", default=False)
        subparser.add_argument('--settings', nargs=1)
        subparser.add_argument('--pythonpath', nargs=1)
        subparser.add_argument('--no-color', action="store_true", default=False)

        return subparser



#-------------------------------------------------------------------------------
#
# WPS 1.0 base XML encoder
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from eoxserver.core.util.xmltools import XMLEncoder
from eoxserver.services.ows.wps.v10.util import ns_wps

class WPS10BaseXMLEncoder(XMLEncoder):
    """ Base class of the WPS 1.0 XML response encoders. """
    content_type = "application/xml; charset=utf-8"

    def get_schema_locations(self):
        return {"wps": ns_wps.schema_location}

    def serialize(self, tree, **kwargs):
        """ Serialize a XML tree to the pair (tuple) of the XML string
        and the content type.
        """
        # override the default ASCII encoding to the standard UTF-8
        kwargs['encoding'] = 'utf-8'
        payload = super(WPS10BaseXMLEncoder, self).serialize(tree, **kwargs)
        return payload, self.content_type



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2015 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


from itertools import chain

from lxml.builder import ElementMaker
try:
    from django.core.urlresolvers import reverse
except ImportError:
    from django.urls import reverse
from django.shortcuts import get_object_or_404

from eoxserver.core.config import get_eoxserver_config
from eoxserver.core.util.xmltools import (
    XMLEncoder, NameSpace, NameSpaceMap
)
from eoxserver.resources.coverages import models
from eoxserver.services.opensearch.formats import get_formats
from eoxserver.services.opensearch.extensions import get_extensions
from eoxserver.services.opensearch.config import OpenSearchConfigReader


class OpenSearch11DescriptionEncoder(XMLEncoder):
    content_type = "application/opensearchdescription+xml"

    def __init__(self, search_extensions):
        ns_os = NameSpace("http://a9.com/-/spec/opensearch/1.1/", None)
        self.ns_param = ns_param = NameSpace(
            "http://a9.com/-/spec/opensearch/extensions/parameters/1.0/",
            "parameters"
        )
        ns_atom = NameSpace("http://www.w3.org/2005/Atom", "atom")
        nsmap = NameSpaceMap(ns_os, ns_param, ns_atom)
        for search_extension in search_extensions:
            nsmap.add(search_extension.namespace)
        self.OS = ElementMaker(namespace=ns_os.uri, nsmap=nsmap)
        self.PARAM = ElementMaker(namespace=ns_param.uri, nsmap=nsmap)
        self.ATOM = ElementMaker(namespace=ns_atom.uri, nsmap=nsmap)
        self.search_extensions = search_extensions

    def encode_description(self, request, collection, result_formats):
        """ Encode an OpenSearch 1.1 description document.
        """
        OS = self.OS
        description = OS("OpenSearchDescription",
            OS("ShortName",
                collection.identifier if collection is not None else ""
            ),
            OS("Description")
        )
        for method in ("GET", "POST"):
            description.extend([
                self.encode_url(
                    request, collection, result_format, method
                )
                for result_format in result_formats
            ])
        description.extend([
            OS("Contact"),
            OS("Tags", "CEOS-OS-BP-V1.1/L1"),
            OS("LongName"),
            OS("Developer"),
            OS("Attribution"),
            OS("SyndicationRight", "open"),
            OS("AdultContent"),
            OS("Language"),
            OS("InputEncoding"),
            OS("OutputEncoding")
        ])
        return description

    def encode_url(self, request, collection, result_format, method):
        """ Encode a single opensearch URL, either for a specific collection, or
            the whole service.
        """
        if collection is not None:
            search_url = reverse("opensearch:collection:search",
                kwargs={
                    "collection_id": collection.identifier,
                    "format_name": result_format.name
                }
            )
        else:
            search_url = reverse("opensearch:search",
                kwargs={
                    "format_name": result_format.name
                }
            )

        conf = OpenSearchConfigReader(get_eoxserver_config())
        search_url = request.build_absolute_uri(search_url)

        default_parameters = (
            dict(name="q", type="searchTerms", profiles=[
            ]),
            dict(name="count", type="count", min=0, max=conf.max_count),
            dict(name="startIndex", type="startIndex", min=0),
        )

        parameters = list(chain(default_parameters, *[
            [
                dict(parameter, **{"namespace": search_extension.namespace})
                for parameter in search_extension.get_schema(
                    collection,
                    models.Collection if collection is None else models.Product
                )
            ] for search_extension in self.search_extensions
        ]))

        query_template = "&".join(
            "%s={%s%s%s%s}" % (
                parameter["name"],
                parameter["namespace"].prefix
                if "namespace" in parameter else "",
                ":" if "namespace" in parameter else "",
                parameter["type"],
                "?" if parameter.get("optional", True) else ""
            )
            for parameter in parameters
        )

        url = self.OS("Url", *[
                self.encode_parameter(parameter, parameter.get("namespace"))
                for parameter in parameters
            ],
            type=result_format.mimetype,
            template="%s?%s" % (search_url, query_template)
            if method == "GET" else search_url,
            rel="results" if collection is not None else "collection", ** {
                self.ns_param("method"): method,
                self.ns_param("enctype"): "application/x-www-form-urlencoded",
                "indexOffset": "0"
            }
        )

        return url

    def encode_parameter(self, parameter, namespace):
        options = parameter.pop("options", [])
        profiles = parameter.pop("profiles", [])

        attributes = {"name": parameter["name"]}
        if namespace:
            attributes["value"] = "{%s:%s}" % (
                namespace.prefix, parameter.pop("type")
            )
        else:
            attributes["value"] = "{%s}" % parameter.pop("type")

        if 'min' in parameter:
            attributes['minInclusive'] = str(parameter['min'])

        if 'max' in parameter:
            attributes['maxInclusive'] = str(parameter['max'])

        pattern = parameter.get("pattern")
        if pattern:
            attributes["pattern"] = pattern

        return self.PARAM("Parameter", *[
            self.PARAM("Option", value=option, label=option)
            for option in options
        ] + [
            self.ATOM("link",
                rel="profile", href=profile["href"], title=profile["title"]
            )
            for profile in profiles
        ], minimum="0" if parameter.get("optional", True) else "1", maximum="1",
            **attributes
        )


class OpenSearch11DescriptionHandler(object):
    def handle(self, request, collection_id=None):
        collection = None
        if collection_id:
            collection = get_object_or_404(models.Collection,
                identifier=collection_id
            )

        encoder = OpenSearch11DescriptionEncoder([
            extension() for extension in get_extensions()
        ])
        return (
            encoder.serialize(
                encoder.encode_description(
                    request, collection, [format_() for format_ in get_formats()]
                )
            ),
            encoder.content_type
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-06-04 14:26
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('coverages', '0005_eo_object_identifier_validator'),
        ('services', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='servicevisibility',
            name='eo_object',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='service_visibility', to='coverages.EOObject'),
        ),
        migrations.AlterField(
            model_name='servicevisibility',
            name='service',
            field=models.CharField(choices=[(b'wms', b'WMS'), (b'wcs', b'WCS'), (b'os', b'OpenSearch'), (b'wc', b'WebClient')], max_length=4),
        ),
        migrations.AlterUniqueTogether(
            name='servicevisibility',
            unique_together=set([('eo_object', 'service')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.3 on 2017-09-06 19:09
from __future__ import unicode_literals

import django.contrib.gis.db.models.fields
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import re


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('backends', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='AcquisitionStation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(db_index=True, max_length=256, unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='AcquisitionSubType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(db_index=True, max_length=256, unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='AllowedValueRange',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('start', models.FloatField()),
                ('end', models.FloatField()),
            ],
        ),
        migrations.CreateModel(
            name='ArchivingCenter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(db_index=True, max_length=256, unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ArrayDataItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('location', models.CharField(max_length=1024)),
                ('format', models.CharField(blank=True, max_length=64, null=True)),
                ('field_index', models.PositiveSmallIntegerField(default=0)),
                ('band_count', models.PositiveSmallIntegerField(default=1)),
                ('subdataset_type', models.CharField(blank=True, max_length=64, null=True)),
                ('subdataset_locator', models.CharField(blank=True, max_length=1024, null=True)),
                ('bands_interpretation', models.PositiveSmallIntegerField(choices=[(0, b'fields'), (1, b'dimension')], default=0)),
            ],
        ),
        migrations.CreateModel(
            name='Browse',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('location', models.CharField(max_length=1024)),
                ('format', models.CharField(blank=True, max_length=64, null=True)),
                ('style', models.CharField(blank=True, max_length=256, null=True)),
                ('coordinate_reference_system', models.TextField()),
                ('min_x', models.FloatField()),
                ('min_y', models.FloatField()),
                ('max_x', models.FloatField()),
                ('max_y', models.FloatField()),
                ('width', models.PositiveIntegerField()),
                ('height', models.PositiveIntegerField()),
            ],
        ),
        migrations.CreateModel(
            name='BrowseType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=256, validators=[django.core.validators.RegexValidator(re.compile(b'^[a-zA-z_][a-zA-Z0-9_]*$'), message=b'This field must contain a valid Name.')])),
                ('red_or_grey_expression', models.CharField(blank=True, max_length=512, null=True)),
                ('green_expression', models.CharField(blank=True, max_length=512, null=True)),
                ('blue_expression', models.CharField(blank=True, max_length=512, null=True)),
                ('alpha_expression', models.CharField(blank=True, max_length=512, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='CollectionMetadata',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('product_type', models.CharField(blank=True, db_index=True, max_length=256, null=True)),
                ('doi', models.CharField(blank=True, db_index=True, max_length=256, null=True)),
                ('platform', models.CharField(blank=True, db_index=True, max_length=256, null=True)),
                ('platform_serial_identifier', models.CharField(blank=True, db_index=True, max_length=256, null=True)),
                ('instrument', models.CharField(blank=True, db_index=True, max_length=256, null=True)),
                ('sensor_type', models.CharField(blank=True, db_index=True, max_length=256, null=True)),
                ('composite_type', models.CharField(blank=True, db_index=True, max_length=256, null=True)),
                ('processing_level', models.CharField(blank=True, db_index=True, max_length=256, null=True)),
                ('orbit_type', models.CharField(blank=True, db_index=True, max_length=256, null=True)),
                ('spectral_range', models.CharField(blank=True, db_index=True, max_length=256, null=True)),
                ('wavelength', models.IntegerField(blank=True, db_index=True, null=True)),
                ('product_metadata_summary', models.TextField(blank=True, null=True)),
                ('coverage_metadata_summary', models.TextField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='CollectionType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=512, unique=True, validators=[django.core.validators.RegexValidator(re.compile(b'^[a-zA-z_][a-zA-Z0-9_]*$'), message=b'This field must contain a valid Name.')])),
            ],
        ),
        migrations.CreateModel(
            name='CoverageMetadata',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
        ),
        migrations.CreateModel(
            name='CoverageType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=512, unique=True, validators=[django.core.validators.RegexValidator(re.compile(b'^[a-zA-z_][a-zA-Z0-9_]*$'), message=b'This field must contain a valid Name.')])),
            ],
        ),
        migrations.CreateModel(
            name='EOObject',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('identifier', models.CharField(max_length=256, unique=True, validators=[django.core.validators.RegexValidator(re.compile(b'^[a-zA-z_][a-zA-Z0-9_.-]*$'), message=b'This field must contain a valid NCName.')])),
                ('begin_time', models.DateTimeField(blank=True, null=True)),
                ('end_time', models.DateTimeField(blank=True, null=True)),
                ('footprint', django.contrib.gis.db.models.fields.GeometryField(blank=True, null=True, srid=4326)),
                ('inserted', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name='FieldType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('index', models.PositiveSmallIntegerField()),
                ('identifier', models.CharField(max_length=512, validators=[django.core.validators.RegexValidator(re.compile(b'^[a-zA-z_][a-zA-Z0-9_.-]*$'), message=b'This field must contain a valid NCName.')])),
                ('description', models.TextField(blank=True, null=True)),
                ('definition', models.CharField(blank=True, max_length=512, null=True)),
                ('unit_of_measure', models.CharField(blank=True, max_length=64, null=True)),
                ('wavelength', models.FloatField(blank=True, null=True)),
                ('significant_figures', models.PositiveSmallIntegerField(blank=True, null=True)),
                ('numbits', models.PositiveSmallIntegerField(blank=True, null=True)),
                ('signed', models.BooleanField(default=True)),
                ('is_float', models.BooleanField(default=False)),
                ('coverage_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='field_types', to='coverages.CoverageType')),
            ],
            options={
                'ordering': ('index',),
            },
        ),
        migrations.CreateModel(
            name='Frame',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(db_index=True, max_length=256, unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Grid',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=256, null=True, unique=True, validators=[django.core.validators.RegexValidator(re.compile(b'^[a-zA-z_][a-zA-Z0-9_]*$'), message=b'This field must contain a valid Name.')])),
                ('coordinate_reference_system', models.TextField()),
                ('axis_1_name', models.CharField(max_length=256)),
                ('axis_2_name', models.CharField(blank=True, max_length=256, null=True)),
                ('axis_3_name', models.CharField(blank=True, max_length=256, null=True)),
                ('axis_4_name', models.CharField(blank=True, max_length=256, null=True)),
                ('axis_1_type', models.SmallIntegerField(choices=[(0, b'spatial'), (1, b'elevation'), (2, b'temporal'), (3, b'other')])),
                ('axis_2_type', models.SmallIntegerField(blank=True, choices=[(0, b'spatial'), (1, b'elevation'), (2, b'temporal'), (3, b'other')], null=True)),
                ('axis_3_type', models.SmallIntegerField(blank=True, choices=[(0, b'spatial'), (1, b'elevation'), (2, b'temporal'), (3, b'other')], null=True)),
                ('axis_4_type', models.SmallIntegerField(blank=True, choices=[(0, b'spatial'), (1, b'elevation'), (2, b'temporal'), (3, b'other')], null=True)),
                ('axis_1_offset', models.CharField(blank=True, max_length=256, null=True)),
                ('axis_2_offset', models.CharField(blank=True, max_length=256, null=True)),
                ('axis_3_offset', models.CharField(blank=True, max_length=256, null=True)),
                ('axis_4_offset', models.CharField(blank=True, max_length=256, null=True)),
                ('resolution', models.PositiveIntegerField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Mask',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('location', models.CharField(max_length=1024)),
                ('format', models.CharField(blank=True, max_length=64, null=True)),
                ('geometry', django.contrib.gis.db.models.fields.GeometryField(blank=True, null=True, srid=4326)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='MaskType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=512, validators=[django.core.validators.RegexValidator(re.compile(b'^[a-zA-z_][a-zA-Z0-9_]*$'), message=b'This field must contain a valid Name.')])),
            ],
        ),
        migrations.CreateModel(
            name='MetaDataItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('location', models.CharField(max_length=1024)),
                ('format', models.CharField(blank=True, max_length=64, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='NilValue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(max_length=512)),
                ('reason', models.CharField(choices=[(b'http://www.opengis.net/def/nil/OGC/0/inapplicable', b'Inapplicable (There is no value)'), (b'http://www.opengis.net/def/nil/OGC/0/missing', b'Missing'), (b'http://www.opengis.net/def/nil/OGC/0/template', b'Template (The value will be available later)'), (b'http://www.opengis.net/def/nil/OGC/0/unknown', b'Unknown'), (b'http://www.opengis.net/def/nil/OGC/0/withheld', b'Withheld (The value is not divulged)'), (b'http://www.opengis.net/def/nil/OGC/0/AboveDetectionRange', b'Above detection range'), (b'http://www.opengis.net/def/nil/OGC/0/BelowDetectionRange', b'Below detection range')], max_length=512)),
                ('field_types', models.ManyToManyField(blank=True, related_name='nil_values', to='coverages.FieldType')),
            ],
        ),
        migrations.CreateModel(
            name='OrbitNumber',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(db_index=True, max_length=256, unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ProcessingCenter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(db_index=True, max_length=256, unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ProcessingMode',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(db_index=True, max_length=256, unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ProcessorName',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(db_index=True, max_length=256, unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ProductMetadata',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('parent_identifier', models.CharField(blank=True, db_index=True, max_length=256, null=True)),
                ('production_status', models.PositiveSmallIntegerField(blank=True, choices=[(0, b'ARCHIVED'), (1, b'ACQUIRED'), (2, b'CANCELLED')], db_index=True, null=True)),
                ('acquisition_type', models.PositiveSmallIntegerField(blank=True, choices=[(0, b'NOMINAL'), (1, b'CALIBRATION'), (2, b'OTHER')], db_index=True, null=True)),
                ('orbit_direction', models.PositiveSmallIntegerField(blank=True, choices=[(0, b'ASCENDING'), (1, b'DESCENDING')], db_index=True, null=True)),
                ('product_quality_status', models.PositiveSmallIntegerField(blank=True, choices=[(0, b'NOMINAL'), (1, b'DEGRAGED')], db_index=True, null=True)),
                ('creation_date', models.DateTimeField(blank=True, db_index=True, null=True)),
                ('modification_date', models.DateTimeField(blank=True, db_index=True, null=True)),
                ('processing_date', models.DateTimeField(blank=True, db_index=True, null=True)),
                ('availability_time', models.DateTimeField(blank=True, db_index=True, null=True)),
                ('start_time_from_ascending_node', models.IntegerField(blank=True, db_index=True, null=True)),
                ('completion_time_from_ascending_node', models.IntegerField(blank=True, db_index=True, null=True)),
                ('illumination_azimuth_angle', models.FloatField(blank=True, db_index=True, null=True)),
                ('illumination_zenith_angle', models.FloatField(blank=True, db_index=True, null=True)),
                ('illumination_elevation_angle', models.FloatField(blank=True, db_index=True, null=True)),
                ('polarisation_mode', models.PositiveSmallIntegerField(blank=True, choices=[(0, b'single'), (1, b'dual'), (2, b'twin'), (3, b'quad'), (4, b'UNDEFINED')], db_index=True, null=True)),
                ('polarization_channels', models.PositiveSmallIntegerField(blank=True, choices=[(0, b'HV'), (1, b'HV, VH'), (2, b'VH'), (3, b'VV'), (4, b'HH, VV'), (5, b'HH, VH'), (6, b'HH, HV'), (7, b'VH, VV'), (8, b'VH, HV'), (9, b'VV, HV'), (10, b'VV, VH'), (11, b'HH'), (12, b'HH, HV, VH, VV'), (13, b'UNDEFINED')], db_index=True, null=True)),
                ('antenna_look_direction', models.PositiveSmallIntegerField(blank=True, choices=[(0, b'LEFT'), (1, b'RIGHT')], db_index=True, null=True)),
                ('minimum_incidence_angle', models.FloatField(blank=True, db_index=True, null=True)),
                ('maximum_incidence_angle', models.FloatField(blank=True, db_index=True, null=True)),
                ('doppler_frequency', models.FloatField(blank=True, db_index=True, null=True)),
                ('incidence_angle_variation', models.FloatField(blank=True, db_index=True, null=True)),
                ('cloud_cover', models.FloatField(blank=True, db_index=True, null=True)),
                ('snow_cover', models.FloatField(blank=True, db_index=True, null=True)),
                ('lowest_location', models.FloatField(blank=True, db_index=True, null=True)),
                ('highest_location', models.FloatField(blank=True, db_index=True, null=True)),
                ('acquisition_station', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='metadatas', to='coverages.AcquisitionStation')),
                ('acquisition_sub_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='metadatas', to='coverages.AcquisitionSubType')),
                ('archiving_center', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='metadatas', to='coverages.ArchivingCenter')),
                ('frame', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='metadatas', to='coverages.Frame')),
                ('orbit_number', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='metadatas', to='coverages.OrbitNumber')),
                ('processing_center', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='metadatas', to='coverages.ProcessingCenter')),
                ('processing_mode', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='metadatas', to='coverages.ProcessingMode')),
                ('processor_name', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='metadatas', to='coverages.ProcessorName')),
            ],
        ),
        migrations.CreateModel(
            name='ProductQualityDegredationTag',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(db_index=True, max_length=256, unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ProductType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=512, unique=True, validators=[django.core.validators.RegexValidator(re.compile(b'^[a-zA-z_][a-zA-Z0-9_]*$'), message=b'This field must contain a valid Name.')])),
                ('allowed_coverage_types', models.ManyToManyField(blank=True, related_name='allowed_product_types', to='coverages.CoverageType')),
            ],
        ),
        migrations.CreateModel(
            name='ProductVersion',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(db_index=True, max_length=256, unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SensorMode',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(db_index=True, max_length=256, unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SwathIdentifier',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(db_index=True, max_length=256, unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Track',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(db_index=True, max_length=256, unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Collection',
            fields=[
                ('eoobject_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='coverages.EOObject')),
            ],
            bases=('coverages.eoobject',),
        ),
        migrations.CreateModel(
            name='Coverage',
            fields=[
                ('eoobject_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='coverages.EOObject')),
                ('axis_1_origin', models.CharField(blank=True, max_length=256, null=True)),
                ('axis_2_origin', models.CharField(blank=True, max_length=256, null=True)),
                ('axis_3_origin', models.CharField(blank=True, max_length=256, null=True)),
                ('axis_4_origin', models.CharField(blank=True, max_length=256, null=True)),
                ('axis_1_size', models.PositiveIntegerField()),
                ('axis_2_size', models.PositiveIntegerField(blank=True, null=True)),
                ('axis_3_size', models.PositiveIntegerField(blank=True, null=True)),
                ('axis_4_size', models.PositiveIntegerField(blank=True, null=True)),
                ('collections', models.ManyToManyField(blank=True, related_name='coverages', to='coverages.Collection')),
                ('coverage_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='coverages', to='coverages.CoverageType')),
                ('grid', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='coverages.Grid')),
            ],
            options={
                'abstract': False,
            },
            bases=('coverages.eoobject', models.Model),
        ),
        migrations.CreateModel(
            name='Mosaic',
            fields=[
                ('eoobject_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='coverages.EOObject')),
                ('axis_1_origin', models.CharField(blank=True, max_length=256, null=True)),
                ('axis_2_origin', models.CharField(blank=True, max_length=256, null=True)),
                ('axis_3_origin', models.CharField(blank=True, max_length=256, null=True)),
                ('axis_4_origin', models.CharField(blank=True, max_length=256, null=True)),
                ('axis_1_size', models.PositiveIntegerField()),
                ('axis_2_size', models.PositiveIntegerField(blank=True, null=True)),
                ('axis_3_size', models.PositiveIntegerField(blank=True, null=True)),
                ('axis_4_size', models.PositiveIntegerField(blank=True, null=True)),
                ('collections', models.ManyToManyField(blank=True, related_name='mosaics', to='coverages.Collection')),
                ('coverage_type', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='mosaics', to='coverages.CoverageType')),
                ('grid', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='coverages.Grid')),
            ],
            options={
                'abstract': False,
            },
            bases=('coverages.eoobject', models.Model),
        ),
        migrations.CreateModel(
            name='Product',
            fields=[
                ('eoobject_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='coverages.EOObject')),
                ('collections', models.ManyToManyField(blank=True, related_name='products', to='coverages.Collection')),
                ('package', models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='backends.Storage')),
                ('product_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='products', to='coverages.ProductType')),
            ],
            bases=('coverages.eoobject',),
        ),
        migrations.CreateModel(
            name='ReservedID',
            fields=[
                ('eoobject_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='coverages.EOObject')),
                ('until', models.DateTimeField(blank=True, null=True)),
                ('request_id', models.CharField(blank=True, max_length=256, null=True)),
            ],
            bases=('coverages.eoobject',),
        ),
        migrations.AddField(
            model_name='productmetadata',
            name='product_quality_degradation_tag',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='metadatas', to='coverages.ProductQualityDegredationTag'),
        ),
        migrations.AddField(
            model_name='productmetadata',
            name='product_version',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='metadatas', to='coverages.ProductVersion'),
        ),
        migrations.AddField(
            model_name='productmetadata',
            name='sensor_mode',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='metadatas', to='coverages.SensorMode'),
        ),
        migrations.AddField(
            model_name='productmetadata',
            name='swath_identifier',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='metadatas', to='coverages.SwathIdentifier'),
        ),
        migrations.AddField(
            model_name='productmetadata',
            name='track',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='metadatas', to='coverages.Track'),
        ),
        migrations.AddField(
            model_name='metadataitem',
            name='eo_object',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='metadata_items', to='coverages.EOObject'),
        ),
        migrations.AddField(
            model_name='metadataitem',
            name='storage',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='backends.Storage'),
        ),
        migrations.AddField(
            model_name='masktype',
            name='product_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='mask_types', to='coverages.ProductType'),
        ),
        migrations.AddField(
            model_name='mask',
            name='mask_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='coverages.MaskType'),
        ),
        migrations.AddField(
            model_name='mask',
            name='storage',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='backends.Storage'),
        ),
        migrations.AddField(
            model_name='collectiontype',
            name='allowed_coverage_types',
            field=models.ManyToManyField(blank=True, related_name='allowed_collection_types', to='coverages.CoverageType'),
        ),
        migrations.AddField(
            model_name='collectiontype',
            name='allowed_product_types',
            field=models.ManyToManyField(blank=True, related_name='allowed_collection_types', to='coverages.ProductType'),
        ),
        migrations.AddField(
            model_name='browsetype',
            name='product_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='browse_types', to='coverages.ProductType'),
        ),
        migrations.AddField(
            model_name='browse',
            name='browse_type',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='coverages.BrowseType'),
        ),
        migrations.AddField(
            model_name='browse',
            name='storage',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='backends.Storage'),
        ),
        migrations.AddField(
            model_name='arraydataitem',
            name='coverage',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='arraydata_items', to='coverages.EOObject'),
        ),
        migrations.AddField(
            model_name='arraydataitem',
            name='storage',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='backends.Storage'),
        ),
        migrations.AddField(
            model_name='allowedvaluerange',
            name='field_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='allowed_value_ranges', to='coverages.FieldType'),
        ),
        migrations.AddField(
            model_name='productmetadata',
            name='product',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='product_metadata', to='coverages.Product'),
        ),
        migrations.AlterUniqueTogether(
            name='masktype',
            unique_together=set([('name', 'product_type')]),
        ),
        migrations.AddField(
            model_name='mask',
            name='product',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='masks', to='coverages.Product'),
        ),
        migrations.AlterUniqueTogether(
            name='fieldtype',
            unique_together=set([('identifier', 'coverage_type'), ('index', 'coverage_type')]),
        ),
        migrations.AddField(
            model_name='coveragemetadata',
            name='coverage',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='coverage_metadata', to='coverages.Coverage'),
        ),
        migrations.AddField(
            model_name='coverage',
            name='mosaics',
            field=models.ManyToManyField(blank=True, related_name='coverages', to='coverages.Mosaic'),
        ),
        migrations.AddField(
            model_name='coverage',
            name='parent_product',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='coverages', to='coverages.Product'),
        ),
        migrations.AddField(
            model_name='collectionmetadata',
            name='collection',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='collection_metadata', to='coverages.Collection'),
        ),
        migrations.AddField(
            model_name='collection',
            name='collection_type',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='collections', to='coverages.CollectionType'),
        ),
        migrations.AddField(
            model_name='collection',
            name='grid',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='coverages.Grid'),
        ),
        migrations.AlterUniqueTogether(
            name='browsetype',
            unique_together=set([('name', 'product_type')]),
        ),
        migrations.AddField(
            model_name='browse',
            name='product',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='browses', to='coverages.Product'),
        ),
        migrations.AlterUniqueTogether(
            name='arraydataitem',
            unique_together=set([('coverage', 'field_index')]),
        ),
        migrations.AlterUniqueTogether(
            name='browse',
            unique_together=set([('product', 'browse_type', 'style')]),
        ),
    ]



from .getcapabilities import WCS10GetCapabilitiesHandler
from .describecoverage import WCS10DescribeCoverageHandler
from .getcoverage import WCS10GetCoverageHandler


GetCapabilitiesHandler = WCS10GetCapabilitiesHandler
DescribeCoverageHandler = WCS10DescribeCoverageHandler
GetCoverageHandler = WCS10GetCoverageHandler



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import re

from django.db.models import ForeignKey
from django.contrib.gis.geos import GEOSGeometry

from eoxserver.contrib import gdal
from eoxserver.backends import models as backends
from eoxserver.backends.storages import get_handler_by_test
from eoxserver.backends.access import get_vsi_path, get_vsi_env
from eoxserver.backends.util import resolve_storage
from eoxserver.resources.coverages import models
from eoxserver.resources.coverages.registration import base
from eoxserver.resources.coverages.metadata.component import (
    ProductMetadataComponent
)
from eoxserver.resources.coverages.registration.exceptions import (
    RegistrationError
)


class ProductRegistrator(base.BaseRegistrator):
    def register(self, metadata_locations, mask_locations, package_path,
                 overrides, identifier_template=None, type_name=None,
                 extended_metadata=True, discover_masks=True,
                 discover_browses=True, discover_metadata=True, replace=False,
                 simplify_footprint_tolerance=None):
        product_type = None
        if type_name:
            product_type = models.ProductType.objects.get(name=type_name)

        component = ProductMetadataComponent()

        browse_handles = []
        mask_locations = mask_locations or []
        metadata = {}

        package = None
        if package_path:
            handler = get_handler_by_test(package_path)
            if not handler:
                raise RegistrationError(
                    'Storage %r is not supported' % package_path
                )

            package, _ = backends.Storage.objects.get_or_create(
                url=package_path, storage_type=handler.name
            )

            if discover_masks or discover_browses or discover_metadata:
                collected_metadata = component.collect_package_metadata(
                    package, handler
                )
                if discover_metadata:
                    metadata.update(collected_metadata)
                if discover_browses:
                    browse_handles.extend([
                        (browse_type, package_path, browse_path)
                        for browse_type, browse_path in metadata.pop(
                            'browses', []
                        )
                    ])
                if discover_masks:
                    mask_locations.extend([
                        (mask_type, package_path, mask_path)
                        for mask_type, mask_path in metadata.pop(
                            'mask_files', []
                        )
                    ])

                    mask_locations.extend([
                        (mask_type, geometry)
                        for mask_type, geometry in metadata.pop('masks', [])
                    ])

        metadata_items = [
            models.MetaDataItem(
                location=location[-1],
                storage=resolve_storage(location[:-1])
            )
            for location in metadata_locations
        ]

        new_metadata = {}
        for metadata_item in reversed(metadata_items):
            new_metadata.update(self._read_product_metadata(
                component, metadata_item
            ))

        mask_locations.extend(new_metadata.pop('masks', []))

        metadata.update(new_metadata)
        metadata.update(dict(
            (key, value) for key, value in overrides.items()
            if value is not None
        ))

        # apply overrides
        identifier = metadata.get('identifier')
        footprint = metadata.get('footprint')
        begin_time = metadata.get('begin_time')
        end_time = metadata.get('end_time')

        if identifier_template:
            identifier = identifier_template.format(metadata)
            metadata['identifier'] = identifier

        if simplify_footprint_tolerance is not None and footprint:
            footprint = footprint.simplify(
                simplify_footprint_tolerance, preserve_topology=True
            )

        replaced = False
        if replace:
            try:
                models.Product.objects.get(identifier=identifier).delete()
                replaced = True
            except models.Product.DoesNotExist:
                pass

        product = models.Product.objects.create(
            identifier=identifier,
            footprint=footprint,
            begin_time=begin_time,
            end_time=end_time,
            product_type=product_type,
            package=package,
        )

        if extended_metadata and metadata:
            create_metadata(product, metadata)

        # register all masks
        for mask_handle in mask_locations:
            geometry = None
            storage = None
            location = ''
            if isinstance(mask_handle[1], GEOSGeometry):
                geometry = GEOSGeometry(mask_handle[1])
            else:
                storage = resolve_storage(mask_handle[1:-1])
                location = mask_handle[-1]

            try:
                mask_type = models.MaskType.objects.get(
                    name=mask_handle[0], product_type=product_type
                )
            except models.MaskType.DoesNotExist:
                raise

            models.Mask.objects.create(
                product=product,
                mask_type=mask_type,
                storage=storage,
                location=location,
                geometry=geometry
            )

        # register all browses
        for browse_handle in browse_handles:
            browse_type = None
            if browse_handle[0]:
                # TODO: only browse types for that product type
                browse_type = models.BrowseType.objects.get(
                    name=browse_handle[0], product_type=product_type
                )

            browse = models.Browse(
                product=product,
                location=browse_handle[-1],
                storage=resolve_storage(browse_handle[1:-1])
            )

            # Get a VSI handle for the browse to get the size, extent and CRS
            # via GDAL
            vsi_path = get_vsi_path(browse)
            ds = gdal.Open(vsi_path)
            browse.width = ds.RasterXSize
            browse.height = ds.RasterYSize
            browse.coordinate_reference_system = ds.GetProjection()
            extent = gdal.get_extent(ds)
            browse.min_x, browse.min_y, browse.max_x, browse.max_y = extent

            browse.full_clean()
            browse.save()

        for metadata_item in metadata_items:
            metadata_item.eo_object = product
            metadata_item.full_clean()
            metadata_item.save()

        return product, replaced

    def _read_product_metadata(self, component, metadata_item):
        path = get_vsi_path(metadata_item)
        with gdal.config_env(get_vsi_env(metadata_item.storage)):
            return component.read_product_metadata_file(path)


def create_metadata(product, metadata_values):
    value_items = [
        (convert_name(name), value)
        for name, value in metadata_values.items()
        if value is not None
    ]

    metadata_values = dict(
        (name, convert_value(name, value, models.ProductMetadata))
        for name, value in value_items
        if value is not None and has_field(models.ProductMetadata, name)
    )

    models.ProductMetadata.objects.create(
        product=product, **metadata_values
    )


def is_common_value(field):
    try:
        if isinstance(field, ForeignKey):
            field.related_model._meta.get_field('value')
            return True
    except:
        pass
    return False


def has_field(model, field_name):
    try:
        model._meta.get_field(field_name)
        return True
    except:
        return False


def camel_to_underscore(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def convert_name(name):
    namespace, _, sub_name = name.partition(':')
    if namespace in ('eop', 'opt', 'sar', 'alt'):
        return camel_to_underscore(sub_name)
    return camel_to_underscore(name)


def convert_value(name, value, model_class):
    field = model_class._meta.get_field(name)
    if is_common_value(field):
        return field.related_model.objects.get_or_create(
            value=value
        )[0]
    elif field.choices:
        return dict((v, k) for k, v in field.choices)[value]
    return value



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


import logging

from eoxserver.core.config import get_eoxserver_config
from eoxserver.core.decoders import config, typelist
from eoxserver.contrib import mapserver as ms
from eoxserver.services.mapserver.wms.layerfactories.base import (
    BaseStyleMixIn, AbstractLayerFactory
)


logger = logging.getLogger(__name__)


class ColorizedMaskLayerFactory(BaseStyleMixIn, AbstractLayerFactory):

    @property
    def suffixes(self):
        return ["_%s" % mask for mask in self.enabled_masks]

    @property
    def enabled_masks(self):
        decoder = EnabledMasksConfigReader(get_eoxserver_config())
        return decoder.mask_names

    def generate(self, eo_object, group_layer, suffix, options):
        mask_name = suffix[1:]
        coverage = eo_object.cast()
        layer_name = "%s%s" % (coverage.identifier, suffix)

        if mask_name not in self.enabled_masks:
            return

        mask_items = coverage.data_items.filter(
            semantic="polygonmask[%s]" % mask_name
        )

        # externaly enforced group or multiple groupped masks
        if group_layer or len(mask_items) > 1:

            # check if the external group provided - if not create a new one
            if not group_layer:
                group_layer = self.generate_group(coverage.identifier + suffix)
                yield group_layer, ()

            # append to a comma separated list
            def _append(l, v):
                return "%s,%s" % (l, v) if l else v

            # handle the groupped layers
            for i, mask_item in enumerate(mask_items):
                layer = self.create_polygon_layer(
                    coverage, "%s_%d" % (layer_name, i)
                )
                group_layer.connection = _append(
                    group_layer.connection, layer.name
                )

                yield layer, (mask_item,)

        # only one mask used directly as a layer
        elif len(mask_items) == 1:
            layer = self.create_polygon_layer(coverage, layer_name)
            self.apply_styles(layer, fill=True)
            yield layer, (mask_items[0],)

    def generate_group(self, name):
        layer = ms.layerObj()
        layer.name = name
        layer.type = ms.MS_LAYER_POLYGON
        layer.connectiontype = ms.MS_UNION
        layer.connection = ""
        self.apply_styles(layer, fill=True)
        return layer

    def create_polygon_layer(self, coverage, name):
        layer = self._create_layer(coverage, name, coverage.extent)
        self._set_projection(layer, coverage.spatial_reference)
        layer.type = ms.MS_LAYER_POLYGON
        layer.dump = True
        layer.offsite = ms.colorObj(0, 0, 0)
        return layer


class EnabledMasksConfigReader(config.Reader):
    section = "services.ows.wms"
    mask_names = config.Option(type=typelist(str, ","), default=[])



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2016 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


from os.path import join, isfile, dirname, abspath
import sys
from ctypes.util import find_library

from django.core.management import call_command

import eoxserver


def create_instance(instance_id, target=None, init_spatialite=False,
                    verbosity=None, traceback=False):
    """ Creates a new EOxServer instance with all necessary files and
        folder structure. Optionally, a SQLite database is initiated.
    """
    # Locate instance template
    instance_template_dir = join(
        dirname(eoxserver.__file__), "instance_template"
    )
    if not isfile(join(instance_template_dir, "manage.py")):
        instance_template_dir = join(sys.prefix, "eoxserver/instance_template")
        if not isfile(join(instance_template_dir,  "manage.py")):
            raise RuntimeError("Error: EOxServer instance template not found.")

    # Add template and extension to options
    options = {
        'template': instance_template_dir,
        'extensions': ["conf", "py"],
        'verbosity': verbosity,
        'traceback': traceback
    }

    args = [instance_id]
    if target is not None:
        args.append(target)

    # create the initial django folder structure
    print("Initializing django project folder.")
    call_command("startproject", *args, **options)

    if init_spatialite:
        _init_spatialite(instance_id, target)


def _init_spatialite(instance_id, target):
    # initialize the spatialite database file
    if target is None:
        dst_data_dir = join(instance_id, instance_id, "data")
        print(dst_data_dir)
    else:
        dst_data_dir = join(abspath(target), instance_id, "data")

    db_name = join(dst_data_dir, "config.sqlite")
    print("Setting up initial database.")
    try:
        from pyspatialite import dbapi2 as db
        conn = db.connect(db_name)
        print("Using pyspatialite.")
    except ImportError:
        try:
            from pysqlite2 import dbapi2 as db
            print("Using pysqlite.")
        except ImportError:
            from sqlite3 import dbapi2 as db
            print("Using sqlite3.")

        conn = db.connect(db_name)
        try:
            conn.enable_load_extension(True)
        except:
            raise Exception(
                "SQLite API does not allow loading of extensions."
            )

        spatialite_lib = find_library('spatialite')
        try:
            print("Trying to load extension module '%s'." % spatialite_lib)
            conn.execute("SELECT load_extension('%s')" % (spatialite_lib,))
        except Exception as msg:
            raise Exception(
                'Unable to load the SpatiaLite library extension '
                '"%s" because: %s' % (spatialite_lib, msg)
            )

    rs = conn.execute('SELECT spatialite_version()')
    rs = rs.fetchone()[0].split(".")
    if (int(rs[0]), int(rs[1])) >= (2, 4):
        print("SpatiaLite found, initializing using "
              "'InitSpatialMetadata()'.")
        # Since spatialite 4.1, InitSpatialMetadata() is not longer run
        # automatically in a transaction. It has to be triggered
        # explicitly.
        if (int(rs[0]), int(rs[1])) >= (4, 1):
            conn.execute("SELECT InitSpatialMetadata(1)")
        else:
            conn.execute("SELECT InitSpatialMetadata()")
    else:
        print("SpatiaLite version <2.4 found, trying to "
              "initialize using 'init_spatialite-2.3.sql'.")
        init_sql_path = "init_spatialite-2.3.sql"
        with open(init_sql_path, 'r') as init_sql_file:
            conn.executescript(init_sql_file.read())
    conn.commit()
    conn.close()



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import sys
from itertools import chain

from django.core.management.base import CommandError, BaseCommand
from django.db.models import Q

from eoxserver.resources.coverages import models
from eoxserver.resources.coverages.management.commands import (
    CommandOutputMixIn, SubParserMixIn
)


class Command(CommandOutputMixIn, SubParserMixIn, BaseCommand):
    def add_arguments(self, parser):
        check_parser = self.add_subparser(parser, 'check')
        list_parser = self.add_subparser(parser, 'list')

        check_parser.add_argument(
            'identifiers', nargs='+',
            help='The identifiers of the objects to check for existence.'
        )
        check_parser.add_argument(
            '-t', '--type', dest="type_name", default="EOObject",
            help=("Optional. Restrict the listed identifiers to given type.")
        )

        list_parser.add_argument(
            'identifiers', nargs='*',
            help=('Optional. The identifiers of the objects to check for '
                  'existence.')
        )
        list_parser.add_argument(
            '-t', '--type', dest="type_name", default="EOObject",
            help=("Optional. Restrict the listed identifiers to given type.")
        )
        list_parser.add_argument(
            '-r', '--recursive',
            dest="recursive", action="store_true", default=False,
            help=("Optional. Recursive listing for collections/products.")
        ),
        list_parser.add_argument(
            '-s', '--suppress-type',
            dest="suppress_type", action="store_true", default=False,
            help=("Optional. Supress the output of the type. By default, the "
                  "type is also printed after the identifier.")
        )
        list_parser.add_argument(
            '-c', '--collection',
            dest="collections", action="append", default=None,
            help=("Optional. Specify a collection that objects must be part "
                  "of. Can be specified multiple times.")
        )

    def handle(self, subcommand, *args, **kwargs):
        if subcommand == "check":
            return self.handle_check(*args, **kwargs)
        elif subcommand == "list":
            return self.handle_list(*args, **kwargs)

    def handle_check(self, identifiers, type_name, *args, **kwargs):
        if not identifiers:
            raise CommandError("Missing the mandatory identifier(s).")

        base_qs = self.get_queryset(type_name)

        if type_name == "EOObject":
            base_qs = base_qs.select_subclasses()

        used = False
        for identifier in identifiers:
            try:
                obj = base_qs.get(identifier=identifier)
                self.print_msg(
                    "The identifier '%s' is already in use by a '%s'."
                    % (identifier, type(obj).__name__)
                )
                used = True
            except base_qs.model.DoesNotExist:
                self.print_msg(
                    "The identifier '%s' is currently not in use." % identifier
                )

        if used:
            sys.exit(1)

    def handle_list(self, identifiers, type_name, suppress_type, collections,
                    **kwargs):
        eo_objects = self.get_queryset(type_name)
        if type_name == 'EOObject':
            eo_objects = eo_objects.select_subclasses()

        if identifiers:
            eo_objects = eo_objects.filter(identifier__in=identifiers)

        if collections:
            eo_objects = eo_objects.filter(
                Q(coverage__collections__identifier__in=collections)
                | Q(product__collections__identifier__in=collections)
            )

        for eo_object in eo_objects:
            self.print_object(eo_object, kwargs["recursive"], suppress_type)

    def get_queryset(self, type_name):
        try:
            # TODO: allow types residing in different apps
            ObjectType = getattr(models, type_name)
            if not issubclass(ObjectType, models.EOObject):
                raise CommandError("Unsupported type '%s'." % type_name)
        except AttributeError:
            raise CommandError("Unsupported type '%s'." % type_name)

        return ObjectType.objects.all()

    def print_object(self, eo_object, recursive=False, suppress_type=False,
                     level=0):
        indent = "  " * level

        if not suppress_type:
            print("%s%s %s" % (indent, eo_object.identifier,
                               eo_object.__class__.__name__))
        else:
            print("%s%s" % (indent, eo_object.identifier))

        if recursive:
            products = []
            coverages = []
            if isinstance(eo_object, models.Collection):
                products = eo_object.products.all()
                coverages = eo_object.coverages.all()

            elif isinstance(eo_object, models.Product):
                coverages = eo_object.coverages.all()

            for sub_eo_object in chain(products, coverages):
                self.print_object(
                    sub_eo_object, recursive, suppress_type, level+1
                )



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Stephan Krause <stephan.krause@eox.at>
#          Stephan Meissl <stephan.meissl@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import logging
import os
import datetime
import httplib
import xml.dom.minidom
import eoxserver
from urlparse import urlparse

from django.utils.six import iteritems

from eoxserver.core import implements
from eoxserver.core.config import get_eoxserver_config
from eoxserver.services.ows.decoders import get_decoder
from eoxserver.services.auth.base import BasePDP, AuthConfigReader
from eoxserver.services.auth.interfaces import PolicyDecisionPointInterface
                                         

logger = logging.getLogger(__name__)


attrib_subject       = "urn:oasis:names:tc:xacml:1.0:subject:subject-id"
attrib_auth_time     = "urn:oasis:names:tc:xacml:1.0:subject:authentication-time"
attrib_auth_method   = "urn:oasis:names:tc:xacml:1.0:subject:authn-locality:authentication-method"
attrib_request_time  = "urn:oasis:names:tc:xacml:1.0:subject:request-time"
attrib_start_time    = "urn:oasis:names:tc:xacml:1.0:subject:session-start-time"
attrib_resource      = "urn:oasis:names:tc:xacml:1.0:resource:resource-id"
attrib_action        = "urn:oasis:names:tc:xacml:1.0:action:action"
attrib_action_ns     = "urn:oasis:names:tc:xacml:1.0:action:action-namespace"
attrib_current_date  = "urn:oasis:names:tc:xacml:1.0:environment:current-dateTime"

dt_string = "http://www.w3.org/2001/XMLSchema#string"
dt_date   = "http://www.w3.org/2001/XMLSchema#dateTime"
dt_any    = "http://www.w3.org/2001/XMLSchema#anyURI"


# Template for the XACMLAuthzDecisionQuery
template_request = "<soapenv:Envelope xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\" " + \
                   "xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\" " + \
                   "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\">" + \
                   "<soapenv:Header></soapenv:Header><soapenv:Body><XACMLAuthzDecisionQuery "  + \
                   "xmlns=\"urn:oasis:xacml:2.0:saml:protocol:schema:os\">" + \
                   "<Request xmlns=\"urn:oasis:names:tc:xacml:2.0:context:schema:os\">" + \
                   "<Subject>{0}</Subject><Resource>{1}</Resource><Action>{2}</Action>" + \
                   "<Environment>{3}</Environment></Request></XACMLAuthzDecisionQuery>" + \
                   "</soapenv:Body></soapenv:Envelope>"

# Template for the Attributes in the XACMLAuthzDecisionQuery
template_attribute = "<Attribute AttributeId=\"{0}\" DataType=\"{1}\">" + \
                     "<AttributeValue>{2}</AttributeValue></Attribute>"


#-------------------------------------------------------------------------------
# PDP implementation for the CHARON Authorization Service
#-------------------------------------------------------------------------------

class CharonPDP(BasePDP):
    implements(PolicyDecisionPointInterface)
    # please do not remove this dictionary; it is needed for EOxServer internal
    # processes

    pdp_type = "charon"

    def __init__(self, client=None):

        cfgReader = AuthConfigReader(get_eoxserver_config())
        url = cfgReader.authorisationServiceURL

        # For tests
        self.client = client or AuthorisationClient(url)
        self.attribMapping = {}

        self.serviceID = cfgReader.serviceID or "default"

        dictLocation = cfgReader.getAttributeMappingDictionary()
        if not dictLocation or dictLocation == "default":
            basePath = os.path.split(eoxserver.__file__)[0]
            dictLocation = os.path.join(basePath, 'conf', 'defaultAttributeDictionary')

        CHAR_COMMENT = '#'
        CHAR_ASSIGN  = '='

        try:
            logger.debug(
                "Loading attribute dictionary from the file %s" % dictLocation
            )
            with open(dictLocation) as f:
                for line in f:
                    if CHAR_COMMENT in line:
                        line, comment = line.split(CHAR_COMMENT, 1)
                    if CHAR_ASSIGN in line:
                        key, value = line.split(CHAR_ASSIGN, 1)
                        key = key.strip()
                        value = value.strip()
                        self.attribMapping[key] = value
                        logger.debug(
                            "Adding SAML attribute to dictionary: %s = %s" 
                            % (key, value)
                        )
        except IOError :
            logger.warn(
                "Cannot read dictionary for attributes mapping from the path: "
                "%s" % dictLocation
            )


    # Extracts the asserted subject attributes from the OWS Request
    def _getAssertedAttributes(self, request):
        httpHeader = request.META
        attributes = {}

        # adding the REMOTE_ADDR from HTTP header to subject attributes
        attributes['REMOTE_ADDR'] = httpHeader['REMOTE_ADDR']

        for key, value in iteritems(self.attribMapping):
            if key in httpHeader:
                attributes[key] = httpHeader[value]
                logger.debug(
                    "Found SAML attribute %s with value %s in incoming "
                    "request." % (key, httpHeader[value]))
            else:
                logger.info(
                    "The key '%s' specified in the mapping dictionary was not "
                    "found in the HTTP headers." % key
                )

        return attributes


    # Extracts the resource specific attributes from the OWS Request
    def _getResourceAttributes(self, request):
        httpHeader = request.META
        attributes = {}

        if self.serviceID == 'default' :
            attributes[attrib_resource] = httpHeader['SERVER_NAME']
        else :
            attributes[attrib_resource] = self.serviceID

        decoder = get_decoder(request)
        attributes['serviceType'] = decoder.service.lower()
        attributes['serverName'] = httpHeader['SERVER_NAME']

        return attributes


    # performs the actual authz. decision
    def _decide(self, request):

        decoder = get_decoder(request)
        userAttributes     = self._getAssertedAttributes(request)
        resourceAttributes = self._getResourceAttributes(request)

        result = self.client.authorize(
            userAttributes, resourceAttributes, decoder.request.lower()
        )
        return result

#-------------------------------------------------------------------------------
# SOAP client for the CHARON Policy Management and Authorization Service
#-------------------------------------------------------------------------------

class AuthorisationClient(object):
    """
    SOAP client for the CHARON Policy Management and Authorisation
    Service

    .. method::  __init__(authz_service_url)

        Constructor with Authorisation Service URL

    .. method:: authorize(userAttributes, resource, action, request)

        This method performs an authorisation request at the Policy Management and
        Authorisation Service.

    """

    def __init__(self, url):

        urlObject = urlparse(url)

        self.port = 80 if urlObject.scheme == 'http' else \
            443 if urlObject.scheme == 'https' else None
        self.hostname = urlObject.hostname
        self.path = urlObject.path if urlObject.path is not None else ''
        self.port = urlObject.port if urlObject.port is not None else self.port
        self.headers = {"Content-type": "text/xml;charset=UTF-8",
                        "Accept": "text/plain",
                        "SOAPAction": "authorise"}

        if (self.hostname is None or self.port is None):
            raise AuthorisationClientException("Invalid argument in constructor: "+str(url)+\
                                               " is not a valid URL.")
        logger.debug("Created instance of AuthorisationClient with the URL "+str(url))


    def authorize(self, userAttributes, resourceAttributes, action):
        request = self._getFullRequest(userAttributes, resourceAttributes, action)

        logger.debug("Sending XACMLAuthzDecisionQuery to "+\
                      str(self.hostname)+":"+str(self.port)+str(self.path)+":\n"+\
                      request)

        connection = httplib.HTTPConnection(self.hostname, self.port)
        connection.request('POST', self.path, request, self.headers)
        response = connection.getresponse()

        # Check for response codes
        if response.status != 200 :
            message = "Received an invalid status code ("+str(response.status)+") when "+\
                      "trying to perform an authorisation query at "+\
                      str(self.hostname)+":"+str(self.port)+str(self.path)+\
                      " Server Message: "+response.reason
            logger.warn(message)
            return (False, message)


        message = response.read()
        connection.close()
        logger.debug("Received the following response from server:\n" + message)
        dom = xml.dom.minidom.parseString(message)
        decision = dom.getElementsByTagNameNS('urn:oasis:names:tc:xacml:2.0:context:schema:os','Decision')

        for node in decision:
            if node.hasChildNodes and node.firstChild.nodeType == xml.dom.minidom.Node.TEXT_NODE:
                value = node.firstChild.data
                if "Permit" == value:
                    return (True, value)
                else :
                    return (False, "Authorisation Decision: "+value)


        return (False, 'Invalid response from Authorisation Service')


    # Get XML snippet for the Subject part of the XACMLAuthzDecisionQuery
    def _getPartSubject(self, userAttributes):
        part = ""
        for attID in userAttributes:
            part += template_attribute.format(attID, \
                                              dt_string, \
                                              userAttributes[attID])
        return part


    # Get XML snippet for the Resource part of the XACMLAuthzDecisionQuery
    def _getPartResource(self, resourceAttributes):

        part = ""
        for attID in resourceAttributes:
            part += template_attribute.format(attID, \
                                              dt_string, \
                                              resourceAttributes[attID])
        return part



    # Get XML snippet for the Action part of the XACMLAuthzDecisionQuery
    def _getPartAction(self, action):
        return template_attribute.format(attrib_action, \
                                         dt_string, \
                                         action)


    # Get XML snippet for the Environment part of the XACMLAuthzDecisionQuery
    def _getPartEnvironment(self):
        now = datetime.datetime.now()
        formattedNow = now.strftime("%Y-%m-%dT%H:%M:%S.%f%z")
        return template_attribute.format(attrib_current_date, \
                                         dt_date, \
                                         formattedNow)


    # Get the full XACMLAuthzDecisionQuery
    def _getFullRequest(self, userAttributes, resourceAttributes, action):
        return template_request.format(self._getPartSubject(userAttributes), \
                                       self._getPartResource(resourceAttributes), \
                                       self._getPartAction(action), \
                                       self._getPartEnvironment())


#-------------------------------------------------------------------------------
# AuthorisationClientException
#-------------------------------------------------------------------------------

class AuthorisationClientException(Exception):
    """ Exception that is thrown by the AuthorisationClient in case of an error
    """



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from django.core.management.base import CommandError, BaseCommand
from django.db import transaction

from eoxserver.resources.coverages import models
from eoxserver.resources.coverages.management.commands import (
    CommandOutputMixIn, SubParserMixIn
)
from eoxserver.resources.coverages.registration.browse import BrowseRegistrator


class Command(CommandOutputMixIn, SubParserMixIn, BaseCommand):
    """ Command to manage browses. This command uses sub-commands for the
        specific tasks: register, generate, deregister
    """
    def add_arguments(self, parser):
        register_parser = self.add_subparser(parser, 'register')
        generate_parser = self.add_subparser(parser, 'generate')
        deregister_parser = self.add_subparser(parser, 'deregister')

        for parser in [register_parser, generate_parser, deregister_parser]:
            parser.add_argument(
                'identifier', nargs=1, help='The associated product identifier'
            )

        register_parser.add_argument(
            'location', nargs='+',
            help="The storage location of the browse."
        )
        register_parser.add_argument(
            '--type', '--browse-type', '-t', dest='type_name', default=None,
            help='The name of the browse type to associate the browse with.'
        )

    @transaction.atomic
    def handle(self, subcommand, identifier, *args, **kwargs):
        """ Dispatch sub-commands: register, deregister.
        """
        identifier = identifier[0]
        if subcommand == "register":
            self.handle_register(identifier, *args, **kwargs)
        elif subcommand == "generate":
            self.handle_generate(identifier, *args, **kwargs)
        elif subcommand == "deregister":
            self.handle_deregister(identifier, *args, **kwargs)

    def handle_register(self, identifier, location, type_name, **kwargs):
        """ Handle the registration of an existing browse.
        """

        BrowseRegistrator().register(
            product_identifier=identifier,
            location=location,
            type_name=type_name
        )
        print(
            'Successfully registered browse (%r) for product %r'
            % (type_name, identifier)
        )

    def handle_generate(self, identifier, **kwargs):
        """ Handle the generation of a new browse image
        """
        raise NotImplementedError

    def handle_deregister(self, identifier, **kwargs):
        """ Handle the deregistration a browse image
        """
        raise NotImplementedError



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------


from lxml.builder import E

from eoxserver.core.util.xmltools import XMLEncoder


class WMS10Encoder(XMLEncoder):
    def encode_capabilities(self, config, ows_url, srss, formats, info_formats,
                            layer_descriptions):

        mime_to_name = {
            "image/gif": "GIF",
            "image/png": "PNG",
            "image/jpeg": "JPEG",
            "image/tiff": "GeoTIFF",

        }

        return E("WMT_MS_Capabilities",
            E("Service",
                E("Name", config.name),
                E("Title", config.title),
                E("Abstract", config.abstract),
                E("Keywords", " ".join(config.keywords)),
                E("OnlineResource", config.onlineresource),
                E("Fees", config.fees),
                E("AccessConstraints", config.access_constraints),
            ),
            E("Capability",
                E("Request",
                    E("Map",
                        E("Format", *[
                                E(mime_to_name[frmt.mimeType])
                                for frmt in formats
                                if frmt.mimeType in mime_to_name
                            ]
                        ),
                        E("DCPType",
                            E("HTTP",
                                E("Get", onlineResource=ows_url)
                            )
                        )
                    ),
                    E("Capabilities",
                        E("Format",
                            E("WMS_XML")
                        ),
                        E("DCPType",
                            E("HTTP",
                                E("Get", onlineResource=ows_url)
                            )
                        )
                    ),
                    E("FeatureInfo",
                        E("Format",
                            # TODO
                        ),
                        E("DCPType",
                            E("HTTP",
                                E("Get", onlineResource=ows_url)
                            )
                        )
                    ),
                ),
                E("Exception",
                    E("Format",
                        E("BLANK"),
                        E("INIMAGE"),
                        E("WMS_XML")
                    ),
                ),
                E("Layer",
                    E("Title", config.title),
                    E("LatLonBoundingBox",
                        minx="-180", miny="-90", maxx="180", maxy="90"
                    ), *([
                        E("SRS", srs)
                        for srs in srss
                    ] + [
                        self.encode_layer(layer_description)
                        for layer_description in layer_descriptions
                    ])
                )
            ),
            version="1.0.0", updateSequence=config.update_sequence
        )

    def encode_layer(self, layer_description):
        elems = [
            E("Name", layer_description.name)
        ]

        title = getattr(layer_description, 'title')
        if title:
            elems.append(E("Title", title))

        if layer_description.bbox:
            bbox = list(map(str, layer_description.bbox))
            elems.append(
                E("LatLonBoundingBox",
                    minx=bbox[0], miny=bbox[1], maxx=bbox[2], maxy=bbox[3]
                )
            )

        elems.extend(
            E("Style",
                E("Name", style),
                E("Abstract", style),
            ) for style in layer_description.styles
        )

        elems.extend(
            self.encode_layer(sub_layer)
            for sub_layer in layer_description.sub_layers
        )

        return E("Layer",
            *elems,
            queryable="1" if layer_description.queryable else "0"
        )



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from django.core.management.base import CommandError, BaseCommand
from django.db import transaction

from eoxserver.backends import models as backends
from eoxserver.backends.storages import (
    get_handler_by_test, get_handler_class_by_name
)
from eoxserver.backends.util import resolve_storage_and_path, resolve_storage
from eoxserver.backends.access import vsi_list_storage, get_vsi_env
from eoxserver.resources.coverages.management.commands import (
    CommandOutputMixIn, SubParserMixIn
)


class Command(CommandOutputMixIn, SubParserMixIn, BaseCommand):
    """ Command to manage storages. This command uses sub-commands for the
        specific tasks: create, delete
    """
    def add_arguments(self, parser):
        create_parser = self.add_subparser(parser, 'create')
        delete_parser = self.add_subparser(parser, 'delete')
        list_parser = self.add_subparser(parser, 'list')
        env_parser = self.add_subparser(parser, 'env')

        # name is a common argument
        for parser in [create_parser, delete_parser, list_parser, env_parser]:
            parser.add_argument(
                'name', nargs=1, help='The storage name'
            )

        create_parser.add_argument(
            'url', nargs=1,
            help='The storage location in a URL format. Mandatory.'
        )
        create_parser.add_argument(
            '--type', '-t', dest='type_name', default=None,
            help='The storage type. Optional. Default is auto-detect the type.'
        )
        create_parser.add_argument(
            '--parent', '-p', dest='parent_name', default=None,
            help='The name of the parent storage. Optional.'
        )
        create_parser.add_argument(
            '--storage-auth', '--storage-auth-name', '--auth', '-a',
            dest='storage_auth_name', default=None,
            help='The name of the storage auth to use. Optional',
        )

        for parser in [list_parser, env_parser]:
            parser.add_argument(
                'paths', nargs='*', default=None,
                help=(
                    'Possible sub-path on that storage. This may include '
                    'specifiers for sub-storages'
                )
            )

        list_parser.add_argument(
            '--pattern', '--glob', '-p', '-g',
            dest='pattern', default=None,
            help='A glob pattern to filter the listing of files. Optional.'
        )

    @transaction.atomic
    def handle(self, subcommand, name, *args, **kwargs):
        """ Dispatch sub-commands: create, delete, insert, exclude, purge.
        """
        name = name[0]
        if subcommand == "create":
            self.handle_create(name, *args, **kwargs)
        elif subcommand == "delete":
            self.handle_delete(name, *args, **kwargs)
        elif subcommand == "list":
            self.handle_list(name, *args, **kwargs)
        elif subcommand == "env":
            self.handle_env(name, *args, **kwargs)

    def handle_create(self, name, url, type_name, parent_name,
                      storage_auth_name, **kwargs):
        """ Handle the creation of a new storage.
        """
        url = url[0]
        parent = None

        if type_name:
            handler = get_handler_class_by_name(type_name)
            if not handler:
                raise CommandError(
                    'Storage type %r is not supported' % type_name
                )
        else:
            handler = get_handler_by_test(url)
            if handler:
                type_name = handler.name
            else:
                raise CommandError(
                    'Could not determine type for storage location %r' % url
                )

        if parent_name:
            try:
                parent = backends.Storage.objects.get(name=parent_name)
            except backends.Storage.DoesNotExist:
                raise CommandError(
                    'No such storage with name %r' % parent_name
                )

        if storage_auth_name:
            storage_auth = backends.StorageAuth.objects.get(
                name=storage_auth_name
            )
        else:
            storage_auth = None

        backends.Storage.objects.create(
            name=name, url=url, storage_type=type_name, parent=parent,
            storage_auth=storage_auth,
        )

        self.print_msg(
            'Successfully created storage %s (%s)' % (
                name or url, type_name
            )
        )

    def handle_delete(self, name, **kwargs):
        """ Handle the deletion of a storage
        """
        try:
            storage = backends.Storage.objects.get(name=name)
        except backends.Storage.DoesNotExist:
            raise CommandError('No such storage with name %r' % name)

        type_name = storage.storage_type
        storage.delete()

        self.print_msg(
            'Successfully deleted storage %s (%s)' % (name, type_name)
        )

    def handle_list(self, name, paths=None, pattern=None, **kwargs):
        storage, path = resolve_storage_and_path(
            [name] + paths or [], save=False
        )
        files = vsi_list_storage(storage, path, pattern)
        for filename in files:
            print(filename)

    def handle_env(self, name, paths, **kwargs):
        storage = resolve_storage([name] + paths or [], save=False)
        for key, value in get_vsi_env(storage).items():
            print('%s="%s"' % (key, value))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.3 on 2017-08-28 10:02
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Storage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('url', models.CharField(max_length=1024)),
                ('storage_type', models.CharField(max_length=32)),
                ('name', models.CharField(max_length=1024, null=True, unique=True)),
                ('parent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='backends.Storage')),
            ],
        ),
    ]



class PolicyDecisionPointInterface(object):
    """ This is the interface for Policy Decision Point (PDP) implementations.
    """

    @property
    def pdp_type(self):
        """ The type name of this PDP.
        """

    def authorize(self, request):
        """ This method takes an :class:`~.OWSRequest` object as input and returns an
            :class:`~.AuthorizationResponse` instance. It is expected to check if
            the authenticated user (if any) is authorized to access the requested
            resource and set the ``authorized`` flag of the response accordingly.

            In case the user is not authorized, the content and status of the
            response shall be filled with an error message and the appropriate
            HTTP Status Code (403).

            The method shall not raise any exceptions.
        """
